
--##########################################################################--
--[[----------------------------------------------------------------------]]--
-- Gradient object
--[[----------------------------------------------------------------------]]--
--##########################################################################--
local lg = love.graphics

local Gradient = {}
Gradient.img = nil
Gradient.img_normal = nil
Gradient.img_reverse = nil
Gradient.angle = 0

local stencil_params = {}
stencil_params.num = 0

local Gradient_mt = { __index = Gradient }
function Gradient:new(gradient_table)
  local Gradient = setmetatable({}, Gradient_mt)
  
  assert(type(gradient_table) == "table", "Error: input must be a table.")
  assert(#gradient_table ~= 0, "Error: gradient table is empty")
  
  is_valid = Gradient:_is_gradient_table_valid(gradient_table)
  assert(is_valid, "Error: gradient table is not valid")
  
  Gradient:_init_img(gradient_table)
  
  return Gradient
end

--##########################################################################--
--[[----------------------------------------------------------------------]]--
-- Public
--[[----------------------------------------------------------------------]]--
--##########################################################################--

--[[  Set the direction of a gradient
    
  ---- Set direction from string 
  
        setDirection(string)
          string can be "up", "down", "left", "right", "upright", "downright", "downleft", "upleft"
    
  ---- Set direction from two strings 
  
        setDirection(string1, string2)
          string1 can be "up", "left", "down", "right"
          string2 can be "up", "left", "down", "right"
    
  ---- Set direction from angle  
  
        setDirection(angle)
          number angle
              direction angle in radians
        
  ---- Set direction from vector 
  
        setDirection(vx, vy)
          number vx
              x component of angle vector
          number vy
              y component of angle vector
        
  ---- Set direction from line
  
        setDirection(x1, y1, x2, y2)
          number x1
              The position of first point on the x-axis.
          number y1
              The position of first point on the y-axis.
          number x2
              The position of second point on the x-axis.
          number y2
              The position of second point on the y-axis.
]]--
function Gradient:setDirection(param1, param2, param3, param4)
  local t1, t2, t3, t4 = type(param1), type(param2), type(param3), type(param4)

  if t1 == "string" and not param2 then
    self:_set_direction_from_string(param1)
  elseif t1 == "string" and t2 == "string" then
    self:_set_direction_from_2string(param1, param2)
  elseif t1 == "number" and not (param2 or param3 or param4) then
    self:_set_direction_from_angle(param1)
  elseif t1 == "number" and t2 == "number" and not (param3 or param4) then
    self:_set_direction_from_vector(param1, param2)
  elseif t1 == "number" and t2 == "number" and t3 == "number" and t4 == "number" then
    self:_set_direction_from_points(param1, param2, param3, param4)
  else
    print("Gradient:setDirection - Warning: could not understand input")
  end
end

--[[  Draw gradient in a rectangle

  number x
      The position of top-left corner along x-axis.
  number y
      The position of top-left corner along y-axis.
  number width
      Width of the rectangle.
  number height
      Height of the rectangle.
]]--
function Gradient:drawRectangle(x, y, width, height)
  if self.angle == 0 or self.angle == 0.5 * math.pi or self.angle == math.pi or 
         self.angle == 1.5*math.pi then
    local scale_x = width / self.img:getWidth()
    local scale_y = height
    lg.draw(self.img, x, y, nil, scale_x, scale_y)
  else
    local w, h = self:_calculate_rotated_rectangle_size(x, y, width, height, self.angle)
    local scale_x = w / self.img:getWidth()
    local scale_y = h
    local cx, cy = x + 0.5*width, y + 0.5*height
    
    stencil_params[1] = x
    stencil_params[2] = y
    stencil_params[3] = width
    stencil_params[4] = height
    
    lg.setStencil(self._rectangle_stencil)
    lg.draw(self.img, cx, cy, self.angle, scale_x, scale_y, 0.5*self.img:getWidth(), 0.5)
    lg.setStencil()
  end
end

--[[ Draw gradient in a circle

    number x
        The position of the center along x-axis.
    number y
        The position of the center along y-axis.
    number radius
        The radius of the circle.
    number segments
        The number of segments used for drawing the circle.
]]--
function Gradient:drawCircle(x, y, radius, segments)
  local w = 2*radius
  local rx, ry = x - radius, y - radius
  
  stencil_params[1] = x
  stencil_params[2] = y
  stencil_params[3] = radius
  stencil_params[4] = segments
  
  lg.setStencil(self._circle_stencil)
  self:_drawRectangleForCircle(rx, ry, w, w)
  lg.setStencil()
end

--[[ Draw gradient in a polygon

  ---- Vertices as a variable number of arguments
  
        number ...
            The vertices of the polygon.
        
  ---- Verticles as a table
  
        table vertices
            The vertices of the polygon as a table.
]]--
function Gradient:drawPolygon(...)
  local t = {...}
  if #t == 1 and type(t[1]) == "table" then
    t = {unpack(t[1])}
  end
  assert(#t >= 6, "Error: polygon must have a minimum of 3 vertices.")
  
  -- find centroid of polygon
  local cx, cy = 0, 0
  for i=1,#t-1,2 do
    local x, y = t[i], t[i+1]
    cx = cx + x
    cy = cy + y
  end
  cx = cx / (0.5*#t)
  cy = cy / (0.5*#t)
  
  -- rotate polygon about centroid by -self.angle
  local s, c = math.sin(-self.angle), math.cos(-self.angle)
  local r = {}
  for i=1,#t-1,2 do
    local x, y = t[i] - cx, t[i+1] - cy
    local xr = x*c - y*s + cx
    local yr = x*s + y*c + cy
    r[i] = xr
    r[i+1] = yr
  end
  
  -- find AABB bounding box of rotated polygon
  local min_x, max_x = math.huge, 0
  local min_y, max_y = math.huge, 0
  for i=1,#r-1,2 do
    local x, y = r[i], r[i+1]
    
    if x < min_x then 
      min_x = x 
    end
    if x > max_x then
      max_x = x
    end
    
    if y < min_y then 
      min_y = y 
    end
    if y > max_y then
      max_y = y
    end
  end
  
  -- Find top left corner and the origin to rotate the rectangle about
  local w = max_x - min_x
  local h = max_y - min_y
  local rx, ry = min_x, min_y
  local rcx, rcy = rx + 0.5*w, ry + 0.5*h
  local rw, rh = (cx - rx) / w, (cy - ry) / h
  
  for i=1,#t do
    stencil_params[i] = t[i]
  end
  stencil_params.num = #t
  
  lg.setStencil(self._polygon_stencil)
  self:_drawRectangleForPolygon(rx, ry, w, h, rw, rh)
  lg.setStencil()
end

--[[  Reverse the order of gradient colours
]]--
function Gradient:reverse()
  if self.img == self.img_normal then
    self.img = self.img_reverse
  else
    self.img = self.img_normal
  end
end

--[[  Set the filter mode for the images

  FilterMode min
      How to scale an image down.
  FilterMode mag
      How to scale an image up.
      
  FilterMode can be "nearest" or "linear"
]]--
function Gradient:setFilter(min, mag)
  self.img_normal:setFilter(min, mag)
  self.img_reverse:setFilter(min, mag)
end

--##########################################################################--
--[[----------------------------------------------------------------------]]--
-- Private
--[[----------------------------------------------------------------------]]--
--##########################################################################--

function Gradient:_is_gradient_table_valid(gradient_table)
  local is_valid = true
  for i=1,#gradient_table do
    local c = gradient_table[i]
    if type(c) ~= "table" then
      is_valid = false
      break
    end
    
    if #c < 4 then
      is_valid = false
      break
    end
    
    if not (type(c[1]) == "number" and type(c[2]) == "number" and
            type(c[3]) == "number" and type(c[4]) == "number") then
       is_valid = false
       break
    end
    
    if not ((c[1] >= 0 and c[1] <= 255) and (c[2] >= 0 and c[2] <= 255) and
            (c[3] >= 0 and c[3] <= 255) and (c[4] >= 0 and c[4] <= 255)) then
        is_valid = false
    end
  end
  
  return is_valid
end

function Gradient:_init_img(g)
  local data = love.image.newImageData(#g, 1)
  local reverse_data = love.image.newImageData(#g, 1)
  for i=1,#g do
    data:setPixel(i-1, 0, g[i][1], g[i][2], g[i][3], g[i][4])
    reverse_data:setPixel(#g-(i-1)-1, 0, g[i][1], g[i][2], g[i][3], g[i][4])
  end
  
  self.img = love.graphics.newImage(data)
  self.img_normal = self.img
  self.img_reverse = love.graphics.newImage(reverse_data)
  
  self.img_normal:setFilter("nearest", "nearest")
  self.img_reverse:setFilter("nearest", "nearest")
end

function Gradient:_set_direction_from_angle(angle)
  self.angle = angle
end

function Gradient:_set_direction_from_vector(vx, vy)
  local lensqr = vx*vx + vy*vy
  if lensqr == 0 then
    return
  end

  self.angle = math.atan2(vy, vx)
end

function Gradient:_set_direction_from_points(p1x, p1y, p2x, p2y)
  self:_set_direction_from_vector(p2x - p1x, p2y - p1y)
end

function Gradient:_set_direction_from_2string(s1, s2)
  if s1 == "up" or s1 == "down" then
    if s2 == "up" or s2 == "down" then
      self:_set_direction_from_string(s1)
    else
      self:_set_direction_from_string(s1..s2)
    end
  else
    if s1 == "up" or s2 == "down" then
      self:_set_direction_from_string(s2)
    else
      self:_set_direction_from_string(s2..s1)
    end
  end
end

function Gradient:_set_direction_from_string(str)
  if     str == "up" then
    self:_set_direction_from_vector(0, -1)
  elseif str == "down" then
    self:_set_direction_from_vector(0, 1)
  elseif str == "left" then
    self:_set_direction_from_vector(-1, 0)
  elseif str == "right" then
    self:_set_direction_from_vector(0, 1)
  elseif str == "upright" then
    self:_set_direction_from_vector(1, -1)
  elseif str == "downright" then
    self:_set_direction_from_vector(1, 1)
  elseif str == "downleft" then
    self:_set_direction_from_vector(-1, 1)
  elseif str == "upleft" then
    self:_set_direction_from_vector(-1, -1)
  end
end

function Gradient:_calculate_rotated_rectangle_size(x, y, width, height, angle)
  local cx, cy = x + 0.5*width, y + 0.5*height
  local s, c = math.sin(angle), math.cos(angle)
  local w, h = 0, 0
  local vx, vy
  
  local x1, y1 = x, y
  vx, vy = x1 - cx, y1 - cy
  x1 = vx * c - vy * s + cx;
  y1 = vx * s + vy * c + cy;
  
  local x2, y2 = x + width, y
  vx, vy = x2 - cx, y2 - cy
  x2 = vx * c - vy * s + cx;
  y2 = vx * s + vy * c + cy;
  
  local x3, y3 = x + width, y + height
  vx, vy = x3 - cx, y3 - cy
  x3 = vx * c - vy * s + cx;
  y3 = vx * s + vy * c + cy;
  
  local x4, y4 = x, y + height
  vx, vy = x4 - cx, y4 - cy
  x4 = vx * c - vy * s + cx;
  y4 = vx * s + vy * c + cy;
  
  local minx = math.min(x1, x2, x3, x4)
  local miny = math.min(y1, y2, y3, y4)
  local maxx = math.max(x1, x2, x3, x4)
  local maxy = math.max(y1, y2, y3, y4)
  local w = math.abs(maxx - minx)
  local h = math.abs(maxy - miny)
  
  return w, h
end

function Gradient:_rectangle_stencil()
  lg.rectangle("fill", stencil_params[1], stencil_params[2], 
                       stencil_params[3], stencil_params[4])
end

function Gradient:_circle_stencil()
  lg.circle("fill", stencil_params[1], stencil_params[2], 
                    stencil_params[3], stencil_params[4])
end

function Gradient:_polygon_stencil()
  local p = {}
  for i=1,stencil_params.num do
    p[i] = stencil_params[i]
  end
  lg.polygon("fill", p)
end

function Gradient:_drawRectangleForCircle(x, y, width, height)
   local scale_x = width / self.img:getWidth()
   local scale_y = height
   local cx, cy = x + 0.5*width, y + 0.5*height
   lg.draw(self.img, cx, cy, self.angle, scale_x, scale_y, 0.5*self.img:getWidth(), 0.5)
end

function Gradient:_drawRectangleForPolygon(x, y, width, height, xratio, yratio)
   local scale_x = width / self.img:getWidth()
   local scale_y = height
   local ox, oy = self.img:getWidth()*xratio, yratio
   local x, y = x + xratio*width, y + yratio*height
   lg.draw(self.img, x, y, self.angle, scale_x, scale_y, ox, oy)
end

return Gradient







